﻿using FliUtils.Mvvm;
using FliProCameraLib;
using System;
using System.Windows.Forms;
using System.IO;
using System.IO.Compression;
using System.Runtime.Serialization.Formatters.Binary;
using FLIPilotSimpleViewModel;

namespace FLIPilotViewModel
{
    public class ThumbnailCapturePanelViewModel : ObservableProperties
    {

        //**************************************************************
        #region Constants

        public const string THUMBNAIL_FILENAME_HINT = "ThumbnailCamera";

        #endregion Constants

        //**************************************************************
        #region Constructor and Initialization

        public ThumbnailCapturePanelViewModel()
        {
            fliProCameraManager.ActiveJobCompleteAction += NotifyActiveJobCompleteActionProc;
        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Public Types

        #endregion

        //**************************************************************
        #region Tool Tip Constants

        private const string TOOL_TIP_IMAGE_FILE = "The Image File to be written to the Camera as a Thumbnail.";

        #endregion Tool Tip Constants

        //**************************************************************
        #region Observable Tool Tip Properties

        public string ImageFileTextBoxToolTip { get { return (String.IsNullOrEmpty(ImageFile) ? TOOL_TIP_IMAGE_FILE : ImageFile); } }
        public string ImageFileToolTip { get { return TOOL_TIP_IMAGE_FILE; } }

        #endregion Observable Tool Tip Properties

        //**************************************************************
        #region Observable Properties

        public string ImageFile
        {
            get { return _ImageFile; }
            set
            {
                if (SetValue(ref _ImageFile, value))
                {
                    OnPropertyChanged("ImageFileTextBoxToolTip");
                }
            }
        }
        private string _ImageFile = String.Empty;

        
        public bool IsSaveThumbnailToCameraFromFileChecked
        {
            get { return _IsSaveThumbnailToCameraFromFileChecked; }
            set { SetValue(ref _IsSaveThumbnailToCameraFromFileChecked, value); }
        }
        private bool _IsSaveThumbnailToCameraFromFileChecked = false;

        #endregion Observable Properties

        //**************************************************************
        #region Private Properties

        private bool RetreivingThumbnail = false;

        #endregion

        //**************************************************************
        #region Public Functions

        public void OpenImageFile()
        {
            using (OpenFileDialog dlg = new OpenFileDialog())
            {
                dlg.Title = "Select File List";
                dlg.Filter = "All Files (*.*)|*.*";
                dlg.CheckFileExists = true;
                dlg.Multiselect = true;
                dlg.InitialDirectory = fliProCameraManager.ToolSettings.OpenFileDirectory;
                dlg.RestoreDirectory = true;
                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    ImageFile = dlg.FileName;
                    fliProCameraManager.ToolSettings.OpenFileDirectory = Path.GetDirectoryName(ImageFile);
                    fliProCameraManager.UpdateToolSettings();
                }
            }
        }

        public void RetrieveThumbnailFromCamera()
        {
            if ((fliProCameraManager.ActiveCamera.Type != FliPro.CameraType.GEN1) &&
                (fliProCameraManager.ActiveCamera.Type != FliPro.CameraType.GEN2))
            {
                MessageBox.Show("Thumbnails can only be retrieved from a camera.");
                return;
            }

            byte[] thumbnail = new byte[FliProCamera.MAX_THUMBNAIL_SIZE];
            int status = fliProCameraManager.ActiveCamera.ReadThumbnail(thumbnail.Length, thumbnail);

            if (status < 0)
            {
                if (status == -7)
                {
                    MessageBox.Show("Thumbnails feature is still under development for this Camera.");
                }
                else
                {
                    MessageBox.Show("Problems encountered retrieving the Thumbnail from the Camera.");
                }
            }
            else
            {
                // Debug code

                //FliFrame fliFrame = GetFileFrame(ImageFile);
                //if (!IsSaveThumbnailToCameraFromFileChecked)
                //{
                //    fliFrame = fliProCameraManager.DisplayCamera.GetCurrentFrame();
                //}

                //thumbnail = fliProCameraManager.DisplayCamera.GetThumbnailBytes(fliFrame);

                // end Debug Code

                // Create a new camera to save the data
                FliFrame thumbnailFliFrame = fliProCameraManager.ActiveCamera.GetThumbnailFrame(thumbnail);
                RetreivingThumbnail = true;
                FrameQueue.InsertFrame(thumbnailFliFrame);
            }
        }

        public void SaveThumbnailToCamera()
        {
            FliFrame fliFrame = null;

            if (IsSaveThumbnailToCameraFromFileChecked)
            {
                if (String.IsNullOrEmpty(ImageFile))
                {
                    MessageBox.Show("A valid Image File name is required when 'Save Image File to Camera as Thumbnail' is checked.");
                }
                else if (!File.Exists(ImageFile))
                {
                    MessageBox.Show("Could not find file:\n\n" + ImageFile);
                }
                else
                {
                    fliFrame = GetFileFrame(ImageFile);
                }
            }
            else
            {
                fliFrame = fliProCameraManager.DisplayCamera.GetCurrentFrame();
            }

            if (fliFrame == null)
            {
                MessageBox.Show("There are no frames captured in this Camera to Save.");
            }
            else
            {
                SaveThumbnailToCamera(fliFrame);
            }
        }

        public void SaveThumbnailToCamera(FliFrame fliFrame)
        {
            byte[] rawData = fliProCameraManager.DisplayCamera.GetThumbnailBytes(fliFrame);

            if (rawData.Length > FliProCamera.MAX_THUMBNAIL_SIZE)
            {
                MessageBox.Show("The data did not compress enough to store as a Thumbnail.");
            }
            else
            {
                int bytesWritten = fliProCameraManager.ActiveCamera.WriteThumbnail(rawData.Length, rawData);
                if (bytesWritten < 0)
                {
                    if (bytesWritten == -7)
                    {
                        MessageBox.Show("Thumbnails feature is still under development for this Camera.");
                    }
                    else
                    {
                        MessageBox.Show("Problems encountered writing Thumbnail to the Camera.");
                    }
                }
                else if (bytesWritten != rawData.Length)
                {
                    MessageBox.Show("Problems encountered writing Thumbnail to the Camera.\nBytes written = " + bytesWritten.ToString() +
                        ", Number of Bytes attempted = " + rawData.Length.ToString());
                }
                else
                {
                    MessageBox.Show("Thumbnail successfully saved to the Camera.");
                }
            }
        }

        public void SetDataContext(ControlPanelViewModel vm)
        {
            cpvm = vm;
        }

        #endregion Public Functions

        //**************************************************************
        #region Private Variables

        private FliFrameQueue FrameQueue = FliFrameQueue.Instance;
        private FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;
        private ControlPanelViewModel cpvm = null;

        #endregion Private Variables


        //**************************************************************
        #region Private Functions

        private FliFrame GetFileFrame(string fileName)
        {
            FliFrame fliFrame = null;

            if (File.Exists(fileName))
            {
                FliProCamera fliProCamera = fliProCameraManager.CreateFromFileCamera(fileName);
                fliProCamera.CopyCameraProperties(fliProCameraManager.ActiveCamera);
                SetHistogramSettings(fliProCamera);
                fliProCamera.ClearFrameList();
                fliFrame = ((FliProCameraFromFile)fliProCamera).ProcessFileData();
                fliFrame.IsCapturedAndProcessed = false;
                fliProCamera.LastFrameQueued = fliFrame;
                fliFrame.ParseRawImageData(fliProCamera, 0);
                fliProCamera.AddFrame(fliFrame, FliProCamera.PurgeActionEnum.ForcePurge);
                fliProCamera.CurrentFrame = fliFrame;
                fliProCamera.CurrentFrameIndex = 1;
            }

            return fliFrame;
        }

        private void NotifyActiveJobCompleteActionProc(object sender, FliProCameraManager.ActiveJobCompleteEventArgs e)
        {
            if (e.Success && fliProCameraManager.IsSaveLatestFrameToCameraChecked && (e.Frame != null) && !RetreivingThumbnail)
            {
                // Write the last frame to the camera
                SaveThumbnailToCamera(e.Frame);
            }
            RetreivingThumbnail = false;
        }

        private void SetHistogramSettings(FliProCamera camera)
        {
            FliProCamera.HistogramSettings histogramSettings = cpvm.GetHistogramSettings();
            camera.AbsoluteBrightestThreshold = histogramSettings.AbsoluteBrightestThreshold;
            camera.AbsoluteDimmestThreshold = histogramSettings.AbsoluteDimmestThreshold;
            camera.BrightestThreshold = histogramSettings.BrightestThreshold;
            camera.ContrastCurvePower = histogramSettings.ContrastCurvePower;
            camera.DimmestThreshold = histogramSettings.DimmestThreshold;
            camera.EdgeDiscardPixels = (short)histogramSettings.EdgeDiscardPixels;
            camera.IsDiscardPixelsChecked = histogramSettings.IsDiscardPixelsChecked;
            camera.IsUseLogScaleChecked = histogramSettings.IsUseLogScaleChecked;
            camera.IsUseAbsoluteThresholdsChecked = histogramSettings.IsUseAbsoluteThresholdsChecked;
            camera.SubtractOffset = histogramSettings.SubtractOffset;
        }

        #endregion Private Functions

    }
}
