﻿using System;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using FliUtils.Mvvm;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Windows;
using System.Windows.Interop;

namespace FliProCameraLib
{
    /// <summary>
    /// This is the lowest level base class for all other classes in
    /// FliProCameraLib. It defines constants, enums, structs and entity classes.
    /// It also extends ObservableProperties to reduce the boilerplate code
    /// required to implement observable properties for MVVM
    /// </summary>
    public class FliPro : ObservableProperties, IDisposable
    {
        //**************************************************************
        #region Limits and Constants

        // Convenence
        public const int KILOBYTES = 1024;
        public const int MEGABYTES = KILOBYTES * 1024;
        public const int GIGABYTES = MEGABYTES * 1024;

        // Maximum number of characters in Device Info strings
        public const int FPRO_USB_STRING_MAXLEN = 256;

        // Maximum number of characters in Device Info strings
        public const int FPRO_DEVICE_MAX_PATH_LENGTH = 1024;

        // Maximum number of characters in Sensor Mode Name strings
        public const int FPRO_SENSOR_MODE_NAME_LENGTH = 32;

        // Maximum number of characters in a serial number
        public const int FENG_SERIAL_NUM_MAX_CHARS = 24;

        // Maximum number of characters in a version string
        public const int FPRO_VERSION_STRING_MAXLEN = 32;

        // Maximum number of simultaneously connected cameras
        public const int FPRO_MAX_CONNECTED_CAMERAS = 16;

        // Default minimum number of frames in the camera's frame pool
        public const int DEFAULT_MIN_FRAME_POOL_COUNT = 10;

        // Default minimum number of frames in the camera's tracking frame pool
        public const int DEFAULT_MIN_TRACKING_FRAME_POOL_COUNT = 10;

        // Maximum number of binning modes available
        public const int FPRO_SENSOR_MAX_BINNING_AVAILABLE = 2048;

        // Number of dummy pixels when enabled
        public const int FPRO_DUMMYPIXELS_NUM = 8;

        // Empty Camera serial number definiton
        public const string EMPTY_SERIAL_NUMBER = "0";

        // Empty Camera model definition
        public const string EMPTY_MODEL = "Unknown";
        public const string MICROLINE_ML50100_MODEL = "MicroLine ML50100";

        // Image size
        public const int MAX_IMAGE_WIDTH = 2048; 
        public const int MAX_IMAGE_HEIGHT = 2048;
        public const int DEFAULT_IMAGE_WIDTH = 2048;  //-=-= FIXME
        public const int DEFAULT_IMAGE_HEIGHT = 2048; //-=-= FIXME

        // Scale factor for Temperature setpoint
        public const int TEMPERATURE_SCALE_FACTOR = 10;

        // Scale factor for Sensor gains
        public const double GAIN_SCALE_FACTOR = 1000.0;

        // Pixel depth in bits
        public const int MAX_PIXEL_DEPTH = 12;
        public const int DEFAULT_MAX_PIXEL_DEPTH = 12;
        public const int DEFAULT_PIXEL_DEPTH = 12;

        // Max Raw Frame buffer size
        public const int MAX_FRAME_METADATA_BYTES   = 156;
        public const int MAX_DUMMY_ROWS             = 4094;
        public const int MAX_DUMMY_PIXELS           = 8;
        public const int MAX_LINE_PIXELS            = (MAX_IMAGE_WIDTH + (2 * MAX_DUMMY_PIXELS));
        public const int MAX_LDR_IMAGE_PIXELS       = (MAX_LINE_PIXELS * MAX_IMAGE_HEIGHT);
        public const int MAX_HDR_IMAGE_PIXELS       = (MAX_LDR_IMAGE_PIXELS * 2);
        public const int MAX_LDR_IMAGE_BYTES        = ((MAX_LDR_IMAGE_PIXELS * MAX_PIXEL_DEPTH) + 7) / 8;
        public const int MAX_HDR_IMAGE_BYTES        = ((MAX_HDR_IMAGE_PIXELS * MAX_PIXEL_DEPTH) + 7) / 8;
        public const int MAX_LDR_RAW_FRAME_BYTES    = (MAX_FRAME_METADATA_BYTES + MAX_LDR_IMAGE_BYTES);
        public const int MAX_HDR_RAW_FRAME_BYTES    = (MAX_FRAME_METADATA_BYTES + MAX_HDR_IMAGE_BYTES);

        // Default pixel size in microns
        public const int DEFAULT_PIXEL_WIDTH = 0;   //-=-= FIXME
        public const int DEFAULT_PIXEL_HEIGHT = 0;  //-=-= FIXME

        // Low level command limits imposed by camera hw/fw implementation
        public const uint RAWCMD_MAX_HDRLENGTH = 16;
        public const uint RAWCMD_MAX_DATALENGTH = 256;
        public const uint RAWCMD_MAX_TOTALLENGTH = RAWCMD_MAX_HDRLENGTH + RAWCMD_MAX_DATALENGTH;

        #endregion Limits and Constants

        //**************************************************************
        #region Enumerated Values

        public enum FPROSTREAMERSTATUS : int
        {
            FPRO_STREAMER_STOPPED_ERROR = -1,
            FPRO_STREAMER_STOPPED = 0,
            FPRO_STREAMER_STREAMING,
        }

        #endregion Enumerated Values

        //**************************************************************
        #region FPGA Register Numbers

        // FPGA Registers - from CFPGADevice.h
        public enum FPROFPGAREG : byte
        {
            FIRMWARE_REV = 0,
            SENSOR_1 = 1,
            SENSOR_2 = 2,
            SENSOR_3 = 3,
            SENSOR_4 = 4,
            SENSOR_5 = 5,
            SENSOR_6 = 6,
            SENSOR_7 = 7,
            SENSOR_8 = 8,
            CONTROL = 9,
            CONFIG = 10,
            RESETWP = 11,
            TEST = 12,
            RESERVED13 = 13,
            FRAME_COUNT = 14,
            EXPOSURE_ROWS = 15,
            TPC_SELECT = 16,
            TBS_TX2 = 17,
            TS_TRTS = 18,
            TAZ_VTZ = 19,
            TAB_CINITB = 20,
            TADR_CINITT = 21,
            TADS = 22,
            TWS = 23,
            SYNC_SYNC = 24,
            RS1_BLS1 = 25,
            RS2_BLS2 = 26,
            RST1_RESET = 27,
            RST2 = 28,
            TINITO1_SAMB1 = 29,
            TINITO2_SAMB2 = 30,
            TINITE1_SAMT1 = 31,
            TINITE2_SAMT2 = 32,
            TSO1_READ1 = 33,
            TSO2_READ2 = 34,
            TSE1_CPC1 = 35,
            TSE2_CPC2 = 36,
            TAC1_TAC1 = 37,
            TAC2_TAC2 = 38,
            TRD1 = 39,
            TRD2 = 40,
            HDR1_HDR1 = 41,
            HDR2 = 42,
            HDR3 = 43,
            TX1_TX1 = 44,
            TX2_TX2 = 45,
            TX3_TX3 = 46,
            ROW_FOT = 47,
            READ_RESETCLOCKCOUNTS = 48,
            START_ROWCOLUMNFRAME = 49,
            STOP_ROWCOLUMNFRAME = 50,
            NUMTRACKPERFRAME = 51,

            SENSOR_READ1 = 118,
            SENSOR_READ2 = 119,
            SENSOR_READ3 = 120,
            SENSOR_READ4 = 121,
            SENSOR_READ5 = 122,
            SENSOR_READ6 = 123,
            SENSOR_READ7 = 124,
            SENSOR_READ8 = 125,
            SYSTEM_STATUS2 = 126,
            SYSTEM_STATUS1 = 127
        }

        #endregion FPGA Register Numbers

        //**************************************************************
        #region FPGA Register Data Flags and Masks

        // CONTROL Register Bit definitions	
        public const uint FPRO_FPGA_CONTROLREG_ENABLECAPTURE = 0x00000001;
        public const uint FPRO_FPGA_CONTROLREG_CANCELCAPTURE = 0x00000008;
        public const uint FPRO_FPGA_CONTROLREG_LEDON = 0x00000080;

        // CONFIG Register definitions
        public const uint FPRO_FPGA_CONFIGREG_IMAGEDATA_ENABLE = 0x00000200;
        public const uint FPRO_FPGA_CONFIGREG_DUMMYPIXEL_ENABLE = 0x00000100;
        public const uint FPRO_FPGA_CONFIGREG_HDRMODE_ENABLE = 0x08000000;
        public const uint FPRO_FPGA_CONFIGREG_LDCMODE_ENABLE = 0x1000000;

        // TEST Register definitions
        public const uint FPRO_FPGA_TESTREG_ENABLETESTPATTERN = 0x02;
        public const uint FPRO_FPGA_TESTREG_FORMAT_ROWTYPE = 0x04;

        // Convenience defintions
        public const uint FPRO_FPGA_WRITEALL_BITMASK = 0xFFFFFFFF;

        #endregion FPGA Register Data Flags and Masks

        //**************************************************************
        #region Enumerations

        // Camera Class
        public enum CameraType
        {
            UNKNOWN = 0,
            GEN1 = 1,
            GEN2 = 2,
            FILE = 10,
            GENERATED = 11
        }

        // Test Image Types - how the image data is formatted
        public enum TESTIMAGEMODE
        {
            ROW = 0,
            COLUMN,
            OFF = 0xff
        }

        // Debug Levels
        public enum DBGLEVEL
        {
            NONE = 0,
            ERROR,
            WARNING,
            INFO,
            DEBUG,
            TRACE
        }

        public enum FPROBLACKADJUSTCHAN
        {
            FPRO_BLACK_ADJUST_CHAN_LDR,
            FPRO_BLACK_ADJUST_CHAN_HDR,
        }

        public enum FPRODEVICETYPE
        {
            FPRO_CAM_DEVICE_TYPE_GSENSE400 = 0x01000400,
            FPRO_CAM_DEVICE_TYPE_GSENSE2020 = 0x01002020,
            FPRO_CAM_DEVICE_TYPE_GSENSE4040 = 0x01004040,
            FPRO_CAM_DEVICE_TYPE_KODAK47051 = 0x02047051,
            FPRO_CAM_DEVICE_TYPE_KODAK29050 = 0x02029050,
            FPRO_CAM_DEVICE_TYPE_DC230_42 = 0x03023042,
            FPRO_CAM_DEVICE_TYPE_DC230_84 = 0x03023084
        }

        // Legacy Camera domain flags
        [Flags]
        public enum DeviceDomain : int
        {
            NONE = 0x0000,
            USB = 0x0002,
            CAMERA = 0x0100,
            FILTERWHEEL = 0x0200,
            USB_CAMERA = USB | CAMERA,
            USB_FILTERWHEEL = USB | FILTERWHEEL
        }

        // Legacy Fan Speed
        public enum FliFanSpeed : int
        {
            OFF = 0,
            ON = unchecked((int)0xffffffff)
        }
        
        // USB Connection Speed
        public enum FProConnection
        {
            FPRO_CONNECTION_USB,
            FPRO_CONNECTION_FIBER
        }

        // USB Connection Speed
        public enum FProUSBSpeed
        {
            FULLSPEED,
            HIGHSPEED,
            SUPERSPEED
        }

        // Auxiliary I/O
        /// @enum  FPROAUXIO
        /// @brief Auxiliary I/O Pins
        ///
        /// The camera makes available auxiliary I/O pins for customer defined use.  This enum simply 
        /// assigns a name for each of the pins to be used in the FPROAuxIO_xxx() set of API calls.
        /// <br><br> Note that different camera models can support different Aux I/O pins.  Consult your 
        /// specific camera documentation for supported pins and physical pin outs.
        ///
        /// @var FPROAUXIO::FPRO_AUXIO_1 
        /// @brief Name for AUX I/O Pin 1
        /// @var FPROAUXIO::FPRO_AUXIO_2 
        /// @brief Name for AUX I/O Pin 2
        /// @var FPROAUXIO::FPRO_AUXIO_3 
        /// @brief Name for AUX I/O Pin 3
        /// @var FPROAUXIO::FPRO_AUXIO_4 
        /// @brief Name for AUX I/O Pin 4
        ///
        public enum FPROAUXIO : int
        {
            FPRO_AUXIO_1 = 0x01,
            FPRO_AUXIO_2 = 0x02,
            FPRO_AUXIO_3 = 0x04,
            FPRO_AUXIO_4 = 0x08,
        }

        /// @enum  FPROAUXIO_DIR
        /// @brief Auxiliary I/O Pin Direction
        ///
        /// The camera makes available auxiliary I/O pins for customer defined use.  The pins can be defined
        /// as inputs or ouputs.  This enum is to be used with the FPROAuxIO_xxx() set of API calls to set the
        /// direction of a given AUX I/O pin.  See #FPROAUXIO for more information.
        /// <br><br> Note that different camera models can support different Aux I/O pins.  Consult your 
        /// specific camera documentation for supported pins and physical pin outs.
        ///
        /// @var FPROAUXIO_DIR::FPRO_AUXIO_DIR_IN 
        /// @brief Set AUX I/O pin as an input with respect to the camera.
        /// @var FPROAUXIO_DIR::FPRO_AUXIO_DIR_OUT 
        /// @brief Set AUX I/O pin as an output with respect to the camera.
        public enum FPROAUXIO_DIR : int
        {
            FPRO_AUXIO_DIR_IN = 0,
            FPRO_AUXIO_DIR_OUT,
        }

        /// @enum  FPROAUXIO_STATE
        /// @brief Auxiliary Output State 
        ///
        /// The camera makes available auxiliary I/O pins for customer defined use.  The pins can be defined
        /// as inputs or ouputs.  For pins defined as outputs, this enum is to be used with the FPROAuxIO_xxx() 
        /// set of API calls to set the state of that pin.  See #FPROAUXIO for more information.
        /// <br><br> Note that different camera models can support different Aux I/O pins.  Consult your 
        /// specific camera documentation for supported pins and physical pin outs.
        ///
        /// @var FPROAUXIO_STATE::FPRO_AUXIO_STATE_LOW 
        /// @brief Pin is in the low state.
        /// @var FPROAUXIO_STATE::FPRO_AUXIO_STATE_HIGH 
        /// @brief Pin is in the high state.
        public enum FPROAUXIO_STATE : int
        {
            FPRO_AUXIO_STATE_LOW,
            FPRO_AUXIO_STATE_HIGH,
        }

        /// @enum  FPROAUXIO_EXPACTIVETYPE
        /// @brief Exposure Active Auxiliary Output
        ///
        /// The camera makes available an auxiliary output pin that signals when an exposure is active.  This
        /// enum defines the the set of signal types that may be configured for the output.  Consult your
        /// specific camera documentation for the timing details of each of these signal types.
        /// <br><br> Note that different camera models can support different Aux I/O pins.  Consult your 
        /// specific camera documentation for supported pins and physical pin outs.
        ///
        /// @var FPROAUXIO_EXPACTIVETYPE::FPRO_AUXIO_EXPTYPE_EXPOSURE_ACTIVE 
        /// @brief Exposure Active- consult your camera documentation for timing details.
        /// @var FPROAUXIO_EXPACTIVETYPE::FPRO_AUXIO_EXPTYPE_GLOBAL_EXPOSURE_ACTIVE 
        /// @brief Global Exposure Active- consult your camera documentation for timing details.
        /// @var FPROAUXIO_EXPACTIVETYPE::FPRO_AUXIO_EXPTYPE_FIRST_ROW_SYNC 
        /// @brief First Row Sync- consult your camera documentation for timing details.
        public enum FPROAUXIO_EXPACTIVETYPE : int
        {
            FPRO_AUXIO_EXPTYPE_EXPOSURE_ACTIVE = 0,
            FPRO_AUXIO_EXPTYPE_GLOBAL_EXPOSURE_ACTIVE,
            FPRO_AUXIO_EXPTYPE_FIRST_ROW_SYNC,
            FPRO_AUXIO_EXPTYPE_RESERVED
        }

        #endregion Enumerations

        //**************************************************************
        #region Data Structures

        // Device information 
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct FPRODEVICEINFO
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_USB_STRING_MAXLEN)]
            public string cFriendlyName;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_USB_STRING_MAXLEN)]
            public string cSerialNo;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_DEVICE_MAX_PATH_LENGTH)]
            public string cDevicePath;
            [MarshalAs(UnmanagedType.I4)]
            public FProConnection eConnType;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiVendorId;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiProdId;
            [MarshalAs(UnmanagedType.I4)]
            public FProUSBSpeed USBSpeed;

            public override string ToString()
            {
                StringBuilder str = new StringBuilder();
                str.Append(cFriendlyName);
                str.Append( ", VID " + uiVendorId.ToString("X4"));
                str.Append( ", PID " + uiProdId.ToString("X4"));
                str.Append( ", SN " + cSerialNo);
                return str.ToString(); 
            }
        }

        // Version Information
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct FPRODEVICEVERSIONS
        {
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_VERSION_STRING_MAXLEN)]
            public string FirmwareVersion;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_VERSION_STRING_MAXLEN)]
            public string FPGAVersion;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_VERSION_STRING_MAXLEN)]
            public string ControllerVersion;
        }

        // Legacy Device identifier information returned by libfli.DeviceList
        public struct FliDeviceId
        {
            public DeviceDomain Domain;
            public string Name;
            public string Model;

            public override string ToString()
            {
                return Domain.ToString() + ": " + Name + ", \"" + Model + "\"";
            }
        }

        // Sensor Mode description
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct FPROSTREAMSTATS
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumFrames;
            [MarshalAs(UnmanagedType.U8)]
            public ulong uiTotalBytes;
            [MarshalAs(UnmanagedType.R8)]
            public double dblDiskAvgMBPerSec;
            [MarshalAs(UnmanagedType.R8)]
            public double dblDiskPeakMBPerSec;
            [MarshalAs(UnmanagedType.I4)]
            public FPROSTREAMERSTATUS iStatus;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiReserved;
        }

        // Sensor Mode description
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct SENSEMODE
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint Index;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FPRO_SENSOR_MODE_NAME_LENGTH)]
            public string Name;

            public SENSEMODE(uint index, string name)
            {
                Index = index;
                Name = name;
            }
        }

        // Camera Capabilities Definitions
        [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Unicode)]
        public unsafe struct CAMERACAP
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint Size;                      // Size of this structure (including uiSize)
            [MarshalAs(UnmanagedType.U4)]
            public uint CapVersion;                // Version of this structure
            [MarshalAs(UnmanagedType.U4)]
            public uint DeviceType;                // Device Type- offering more capability differentiation (see FPRODEVICETYPE for values)
            [MarshalAs(UnmanagedType.U4)]
            public uint MaxPixelImageWidth;        // Max allowed image width in pixels
            [MarshalAs(UnmanagedType.U4)]
            public uint MaxPixelImageHeight;       // Max allowed image height in pixels
            [MarshalAs(UnmanagedType.U4)]
            public uint AvailablePixelDepths;      // 1 bit is set for each supported pixel depth (lsb= pixel depth 1)
            [MarshalAs(UnmanagedType.U4)]
            public uint BinRatioTableSize;         // 0 = 1:1 binning only, >0 = number of entries in BinningsTable
            [MarshalAs(UnmanagedType.U4)]
            public uint BlackLevelMax;             // Max value allowed for FPROSensor_SetBlackLevelAdjust() call
            [MarshalAs(UnmanagedType.U4)]
            public uint BlackSunMax;               // Max value allowed for FPROSensor_SetBlackSunAdjust() call
            [MarshalAs(UnmanagedType.U4)]
            public uint LowGain;                   // Number of entries in the LDR Gain table
            [MarshalAs(UnmanagedType.U4)]
            public uint HighGain;                   // Number of entries in the HDR Gain table
            [MarshalAs(UnmanagedType.U4)]
            public uint GlobalGain;                // Number of bits
            [MarshalAs(UnmanagedType.U4)]
            public uint RowScanTime;               // Row scan time in nSec
            [MarshalAs(UnmanagedType.U4)]
            public uint DummyPixelNum;             // Number of Dummy Pixels at the end of a row
            [MarshalAs(UnmanagedType.U1)]
            public bool HScanInvertable;           // False= Normal scan direction only, True= Inverse Scan Available
            [MarshalAs(UnmanagedType.U1)]
            public bool VScanInvertable;           // False= Normal scan direction only, True= Inverse Scan Available
            [MarshalAs(UnmanagedType.U4)]
            public uint NVStorageAvailable;        // Number of bytes of Non-Volatile Storage available on the camera
            [MarshalAs(UnmanagedType.U4)]
            public uint PreFrameReferenceRows;     // Number of Pre-Frame Reference rows available
            [MarshalAs(UnmanagedType.U4)]
            public uint PostFrameReferenceRows;    // Number of Post-Frame Reference rows available
            [MarshalAs(UnmanagedType.U4)]
            public uint MetaDataSize;              // Number of bytes used for the pre-frame image meta data
        }


        ///
        /// This type is used for setting and retrieving the Black Level and Black Sun settings for
        /// the camera.
        /// 
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGBLACK
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint eChannel;           // see FPROBLACKADJUSTCHAN
            [MarshalAs(UnmanagedType.U4)]
            public uint uiValue;            // Value for the setting
        }

        //Engineering Camera Config
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGIMAGEDAC
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint uiChannel;         // Channel setting- upper bits include gain index assigned to this setting
            [MarshalAs(UnmanagedType.U4)]
            public uint uiValue;           // Value for the setting
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGREG
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint uiDevice;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiRegister;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiValue;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiMask;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGMODES
        {
            public SENSEMODE modeInfo;

            [MarshalAs(UnmanagedType.U1)]
            public byte bHDREnabled;
            [MarshalAs(UnmanagedType.U1)]
            public byte bLDCEnabled;

            [MarshalAs(UnmanagedType.I4)]
            public int iLowGainIndex;
            [MarshalAs(UnmanagedType.I4)]
            public int iHighGainIndex;
            [MarshalAs(UnmanagedType.I4)]
            public int iGlobalGainIndex;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumBlackLevel;            // Number of Black Level register settings.
            public IntPtr pBlackLevel;                     // Black Level setting for the mode.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumBlackSun;              // Number of Black Sun register settings.
            public IntPtr pBlackSun;                       // Black Sun setting for the mode.

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumDACImage1;             // Number of Image DAC 1 settings- should be one for each gain setting
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumDACImage2;             // Number of Image DAC 2 settings- should be one for each gain setting
            public IntPtr dacImage1;                // On board image DAC settings for the mode. Pointer to MFENG_CFGIMAGEDAC
            public IntPtr dacImage2;                // On board image DAC settings for the mode. Pointer to MFENG_CFGIMAGEDAC

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumHSMIdleTimingRegs;      // Number of Horizontal State Machine Idle Timing settings.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumHSMBinningTimingRegs;   // Number of Horizontal State Machine Binning Timing settings.
            public IntPtr pCfgHSMIdleTimingRegs;     // The HSM Idle timing settings for the mode.
            public IntPtr pCfgHSMBinningTimingRegs;  // The HSM Binning Timing settings for the mode.

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumOtherRegs;              // Number of Other Register settings
            public IntPtr pModeOtherRegs;            // Other Register settings

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumAtoD1Regs;              // Number of AtoD settings for first AtoD.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumAtoD2Regs;              // Number of AtoD settings for second AtoD.
            public IntPtr pCfgSensorAtoD1;           // AtoD settings for first AtoD.
            public IntPtr pCfgSensorAtoD2;           // AtoD settings for second AtoD.
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct CSFENG_CFGMODES
        {
            public SENSEMODE modeInfo;

            public bool bHDREnabled;
            public bool bLDCEnabled;

            [MarshalAs(UnmanagedType.I4)]
            public int iLowGainIndex;
            [MarshalAs(UnmanagedType.I4)]
            public int iHighGainIndex;
            [MarshalAs(UnmanagedType.I4)]
            public int iGlobalGainIndex;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumBlackLevel;            // Number of Black Level register settings.
            public MFENG_CFGBLACK[] pBlackLevel;           // Black Level setting for the mode.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumBlackSun;              // Number of Black Sun register settings.
            public MFENG_CFGBLACK[] pBlackSun;             // Black Sun setting for the mode.

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumDACImage1;             // Number of Image DAC 1 settings- should be one for each gain setting
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumDACImage2;             // Number of Image DAC 2 settings- should be one for each gain setting
            public int[,] dacImage1;                // On board image DAC settings for the mode.
            public int[,] dacImage2;                // On board image DAC settings for the mode.

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumHSMIdleTimingRegs;              // Number of Horizontal State Machine Idle Timing settings.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumHSMBinningTimingRegs;           // Number of Horizontal State Machine Binning Timing settings.
            public MFENG_CFGREG[] pCfgHSMIdleTimingRegs;     // The HSM Idle timing settings for the mode.
            public MFENG_CFGREG[] pCfgHSMBinningTimingRegs;  // The HSM Binning Timing settings for the mode.

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumOtherRegs;               // Number of Other Register settings
            public MFENG_CFGREG[] pModeOtherRegs;     // Other Register settings

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumAtoD1Regs;               // Number of AtoD settings for first AtoD.
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumAtoD2Regs;               // Number of AtoD settings for second AtoD.
            public MFENG_CFGREG[] pCfgSensorAtoD1;    // AtoD settings for first AtoD.
            public MFENG_CFGREG[] pCfgSensorAtoD2;    // AtoD settings for second AtoD.
        }


        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CAMSTARTUP
        {
            [MarshalAs(UnmanagedType.R8)]
            public double dblTemperatureSetPoint;
            [MarshalAs(UnmanagedType.U4)]
            public uint  uiCoolerDAC;
            [MarshalAs(UnmanagedType.U4)]
            public uint  uiFanPercent;
            [MarshalAs(UnmanagedType.Bool)]
            public bool  bFanEnabled;
            [MarshalAs(UnmanagedType.U4)]
            public uint  uiHeaterPercent;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumVSMIdleTimingRegs;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumVSMLongExpTimingRegs;
            public IntPtr pCfgVSMIdleTimingRegs;
            public IntPtr pCfgVSMLongExpTimingRegs;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumOtherRegs;
            public IntPtr pOtherRegs;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct CSFENG_CAMSTARTUP
        {
            [MarshalAs(UnmanagedType.R8)]
            public double dblTemperatureSetPoint;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiCoolerDAC;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiFanPercent;
            [MarshalAs(UnmanagedType.Bool)]
            public bool bFanEnabled;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiHeaterPercent;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumVSMIdleTimingRegs;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumVSMLongExpTimingRegs;
            public MFENG_CFGREG[] pCfgVSMIdleTimingRegs;
            public MFENG_CFGREG[] pCfgVSMLongExpTimingRegs;

            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumOtherRegs;
            public MFENG_CFGREG[] pOtherRegs;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CAMINFO
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint uiDeviceType;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiSerialNumLength;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = FENG_SERIAL_NUM_MAX_CHARS)]
            public string wcSerialNumber;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct CSFENG_CAMINFO
        {
            public uint uiDeviceType;
            public uint uiSerialNumLength;
            public string wcSerialNumber;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGCAM
        {
            public IntPtr pModes;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiNumModes;
            public MFENG_CAMSTARTUP cfgStartup;
            public MFENG_CAMINFO cfgCamInfo;
        }
        public struct CSFENG_CFGCAM
        {
            public CSFENG_CFGMODES[] modes;
            public uint uiNumModes;
            public CSFENG_CAMSTARTUP cfgStartup;
            public CSFENG_CAMINFO cfgCamInfo;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        public struct MFENG_CFGGAIN
        {
            [MarshalAs(UnmanagedType.U4)]
            public uint uiValue;
            [MarshalAs(UnmanagedType.U4)]
            public uint uiDeviceIndex;
        }

        #endregion Data Structures

        //**************************************************************
        #region Constants and Enums
        // Constants and enums derived from libflipro.h #defines

        // Max length of strings received from library
        public const int MAX_STRING_LEN = 64;

        // Invalid or closed device handle
        public const int INVALID_HANDLE = -1;

        public enum Exposure_Hurtz
        {
            hur_12MHz,
            hur_1_7MHz
        }

        // Bayer/Truesense color filter pattern
        // GRBG =

        // | B G
        // | G R    lower left is (1,1)

        // Truesense Sparse Color Filter Pattern (RGBW #1)
        // BGGR =
        // | G P R P
        // | P G P R
        // | B P G P
        // | P B P G  lower left is (1,1)

        // Bayer Color Filter Pattern
        // "The CFA contains 75% green and 25% red and blue photosites."
        // GBGG_GGGR =
        // | G G G R
        // | G B G G  lower left is (1,1)

        public enum FliBayerPattern : int
        {
            NONE        = 0,
            RGGB        = 1,
            GRBG        = 2,
            GBRG        = 3,
            BGGR        = 4,
            GBGG_GGGR   = 5,
            TRUESENSE   = 6
        }

        // Frame types
        [Flags]
        public enum FliFrameCount : int
        {
            SINGLE = 0,                        // Single
            CONTINUOUS = 1,                    // Used to Continuously Grab from the camera
            MULTIPLE = 2,                      // Used to Grab a fixed number of frames
            STREAM_SINGLE = 3,                 // Streaming Single
            STREAM_CONTINUOUS = 4,             // Streaming Continuous Grabs
            STREAM_MULTIPLE = 5                // Streaming Grab a fixed number of frames
        }

        // Frame types
        [Flags]
        public enum FliFrameType : int
        {
            NORMAL = 0,                   // Normal
            FLAT = 100,                   // Flat
            DARK = 1,                     // DarkFrame
            FLOOD       = 2,              // Flush
            RBI_FLUSH   = FLOOD | DARK,   // Continuous
            BIAS        = 4,              // Bias
            TDI_LIGHT   = 5,              // TDI Light
            TDI_DARK    = 6,              // TDI Dark
        }

        /// FPRO_FRAME_TYPE
        /// Image Frame Type 
        ///
        /// The camera has the ability to produce differnt frame types.  The default
        /// frame type is FPRO_FRAMETYPE_NORMAL.  Consult you camera documentation for
        /// the details of each frame type and availability on a given camera model.
        /// 
        /// See #FPROFrame_SetFrameType() and #FPROFrame_GetFrameType().
        ///
        /// FPRO_FRAMETYPE_NORMAL - Normal Frame (default).
        /// FPRO_FRAMETYPE_DARK - Dark Frame.
        /// FPRO_FRAMETYPE_BIAS - Bias Frame.
        /// FPRO_FRAMETYPE_LIGHTFLASH - Light Flash Frame.
        /// FPRO_FRAMETYPE_DARKFLASH - Dark Flash Frame.
        public enum FPRO_FRAME_TYPE
        {
            FPRO_FRAMETYPE_NORMAL = 0,
            FPRO_FRAMETYPE_DARK,
            FPRO_FRAMETYPE_BIAS,
            FPRO_FRAMETYPE_LIGHTFLASH,
            FPRO_FRAMETYPE_DARKFLASH,
        };
            
        // Trigger controls
        public enum FliTriggerMode : int
        {
            EXTERNAL_LOW = 0,
            EXTERNAL_HIGH = 1,
            EXTERNAL_CONTROL_LOW = 2,
            EXTERNAL_CONTROL_HIGH = 3,
            INTERNAL = 4
        }

        // GPS States
        public enum FliGPSState
        {
            FPRO_GPS_NOT_DETECTED = 0,          // GPS unit has not been detected by the camera.
            FPRO_GPS_DETECTED_NO_SAT_LOCK,      // GPS unit has been detected by the camera but the satellite lock has not been made.
            FPRO_GPS_DETECTED_AND_SAT_LOCK      // GPS unit has been detected by the camera and the satellite lock has been made.  
                                                //     This is the only value that will provide accurate results in the Meta Data.
        }

        // Legacy Shutter controls
        public enum FliShutterControl : int
        {
            CLOSE = 0,
            OPEN = 1,
            EXTERNAL_TRIGGER = 2,
            EXTERNAL_TRIGGER_LOW = 2,
            EXTERNAL_TRIGGER_HIGH = 4,
            EXPOSURE_CONTROL = 8
        }

        // Background Flush
        public enum FliBgFlush : int
        {
            STOP = 0,
            START = 1
        }

        // IO 
        [Flags]
        public enum FliIO : int
        {
            P0          = 0x01,
            P1          = 0x02,
            P2          = 0x04,
            P3          = 0x08
        }

        // Pixel Defects
        public enum FliPixelDefect : int
        {
            COLUMN          = 0x00,
            CLUSTER         = 0x10,
            POINT_BRIGHT    = 0x20,
            POINT_DARK      = 0x30
        }

        // Sensor Gain Tables
        public enum GainTable
        {
            LOW             = 0,
            HIGH            = 1
        }

        #endregion Constants and Enums

        //**************************************************************
        #region Data Structures

        public struct FliGainValue
        {
            public uint uiValue;
            public int uiDeviceIndex;
        }

        /// <summary>
        /// Pixel size (W x H) in microns
        /// </summary>
        public struct FliPixelSize
        {
            public double W;
            public double H;

            public FliPixelSize(double w, double h) 
            {
                W = w;
                H = h;
            }

            public override string ToString()
            {
                return "(" + W.ToString("0.###") + " x " + H.ToString("0.###") + ")";
            }
        }

        /// <summary>
        /// [X,Y] coordinate
        /// </summary>
        public struct FliPoint
        {
            public int X;
            public int Y;

            public FliPoint(int x, int y)
            {
                X = x;
                Y = y;
            }

            public override string ToString()
            {
                return "[" + X.ToString() + "," + Y.ToString() + "]";
            }
        }

        /// <summary>
        /// Rectangular size (W x H)
        /// </summary>
        public struct FliSize
        {
            public int W;
            public int H;

            public FliSize(int w, int h)
            {
                W = w;
                H = h;
            }

            public override string ToString()
            {
                return "(" + W.ToString() + "x" + H.ToString() + ")";
            }
        }

        public struct FliRectangle
        {
            public FliPoint UL;
            public FliPoint LR;

            /// <summary>
            /// Construct from upper left and lower right points
            /// </summary>
            public FliRectangle(FliPoint ul, FliPoint lr)
            {
                UL = ul;
                LR = lr;
            }

            /// <summary>
            /// Construct from 4 coordinates
            /// </summary>
            public FliRectangle(int ulx, int uly, int lrx, int lry)
            {
                UL.X = ulx;
                UL.Y = uly;
                LR.X = lrx;
                LR.Y = lry;
            }

            /// <summary>
            /// Construct from upper left point and size
            /// </summary>
            public FliRectangle(FliPoint ul, FliSize wh)
            {
                UL = ul;
                if (wh.W > 0)
                    LR.X = UL.X + wh.W - 1;
                else
                    LR.X = UL.X;
                if (wh.H > 0)
                    LR.Y = UL.Y + wh.H - 1;
                else
                    LR.Y = UL.Y;
            }

            /// <summary>
            /// Construct from FliRectangleWH
            /// </summary>
            public FliRectangle(FliRectangleWH src)
            {
                UL = src.UL;
                if (src.WH.W > 0)
                    LR.X = UL.X + src.WH.W - 1;
                else
                    LR.X = UL.X;
                if (src.WH.H > 0)
                    LR.Y = UL.Y + src.WH.H - 1;
                else
                    LR.Y = UL.Y;
            }

            public override string ToString()
            {
                return UL.ToString() + LR.ToString();
                //return "[" + UL.X.ToString() + "," + UL.Y.ToString() + "]["
                //           + LR.X.ToString() + "," + LR.Y.ToString() + "]";
            }
        }

        public struct FliRectangleWH
        {
            public FliPoint UL;
            public FliSize WH;
            public int W { get { return WH.W; } set { WH.W = value; } }
            public int H { get { return WH.H; } set { WH.H = value; } }

            /// <summary>
            /// Construct from upper left point and size
            /// </summary>
            public FliRectangleWH(FliPoint ul, FliSize wh)
            {
                UL = ul;
                WH = wh;
            }

            /// <summary>
            /// Construct from upper left point, width and height
            /// </summary>
            public FliRectangleWH(FliPoint ul, int w, int h)
            {
                UL = ul;
                WH.W = w;
                WH.H = h;
            }

            /// <summary>
            /// Construct from upper left coordinates, width and height
            /// </summary>
            public FliRectangleWH(int ulx, int uly, int w, int h)
            {
                UL.X = ulx;
                UL.Y = uly;
                WH.W = w;
                WH.H = h;
            }

            /// <summary>
            /// Construct from upper left and lower right points
            /// </summary>
            public FliRectangleWH(FliPoint ul, FliPoint lr)
            {
                UL = ul;
                WH.W = lr.X - UL.X + 1;
                WH.H = lr.Y - UL.Y + 1;
            }

            /// <summary>
            /// Construct from FliRectangle
            /// </summary>
            public FliRectangleWH(FliRectangle src)
            {
                UL = src.UL;
                WH.W = src.LR.X - UL.X + 1;
                WH.H = src.LR.Y - UL.Y + 1;
                if (WH.H < 0)
                {
                    //int i = 1;
                }
            }

            public override string ToString()
            {
                return UL.ToString() + WH.ToString();
            }
        }

        #endregion Data Stuctures

        //**************************************************************
        #region Error Codes

        /// <summary>
        /// libfli error codes
        /// </summary>
        protected enum ErrorNumber : int
        {
            EPERM = 1,
            ENOENT = 2,
            ESRCH = 3,
            EINTR = 4,
            EIO = 5,
            ENXIO = 6,
            E2BIG = 7,
            ENOEXEC = 8,
            EBADF = 9,
            ECHILD = 10,
            EAGAIN = 11,
            ENOMEM = 12,
            EACCES = 13,
            EFAULT = 14,
            ENOTBLK = 15,
            EBUSY = 16,
            EEXIST = 17,
            EXDEV = 18,
            ENODEV = 19,
            ENOTDIR = 20,
            EISDIR = 21,
            EINVAL = 22,
            ENFILE = 23,
            EMFILE = 24,
            ENOTTY = 25,
            EFBIG = 27,
            ENOSPC = 28,
            ESPIPE = 29,
            EROFS = 30,
            EMLINK = 31,
            EPIPE = 32,
            EDOM = 33,
            ERANGE = 34,
            EDEADLCK = 35,
            EDEADLK = 36,
            ERR037 = 37,
            ENAMETOOLONG = 38,
            ENOLCK = 39,
            ENOSYS = 40,
            ENOTEMPTY = 41,
            EILSEQ = 42,
            STRUNCATE = 80,
            EADDRINUSE = 100,
            EADDRNOTAVAIL = 101,
            EAFNOSUPPORT = 102,
            EALREADY = 103,
            EBADMSG = 104,
            ECANCELED = 105,
            ECONNABORTED = 106,
            ECONNREFUSED = 107,
            ECONNRESET = 108,
            EDESTADDRREQ = 109,
            EHOSTUNREACH = 110,
            EIDRM = 111,
            EINPROGRESS = 112,
            EISCONN = 113,
            ELOOP = 114,
            EMSGSIZE = 115,
            ENETDOWN = 116,
            ENETRESET = 117,
            ENETUNREACH = 118,
            ENOBUFS = 119,
            ENODATA = 120,
            ENOLINK = 121,
            ENOMSG = 122,
            ENOPROTOOPT = 123,
            ENOSR = 124,
            ENOSTR = 125,
            ENOTCONN = 126,
            ENOTRECOVERABLE = 127,
            ENOTSOCK = 128,
            ENOTSUP = 129,
            EOPNOTSUPP = 130,
            EOTHER = 131,
            EOVERFLOW = 132,
            EOWNERDEAD = 133,
            EPROTO = 134,
            EPROTONOSUPPORT = 135,
            EPROTOTYPE = 136,
            ETIME = 137,
            ETIMEDOUT = 138,
            ETXTBSY = 139,
            EWOULDBLOCK = 140
        }

        #endregion Error Codes

        //**************************************************************
        #region Constructor/Destructor

        //-------------------------------------------
        /// <summary>
        /// Default Constructor/Destructor
        /// </summary>
        protected FliPro()
        {
            // Don't do these here anymore
            //DebugEnable = false;
            //DebugLevel = DBGLEVEL.INFO;
        }

        ~FliPro()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing).
            Dispose(false);
        }

        #endregion Constructor/Destructor

        //**************************************************************
        #region Public Properties

        //-------------------------------------------
        /// <summary>
        /// Get the current libfliPro version as a string
        /// (FliGetLibVersion)
        /// </summary>
        public string LibVersion
        {
            get
            {
                var ver = new StringBuilder(FliPro.MAX_STRING_LEN);
                ver.Append("0");
                ThrowOnError(NativeMethods.FPROCam_GetAPIVersion(ver, (uint)ver.MaxCapacity));
                return ver.ToString();
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Get/Set the libfliPro debug log file path
        /// (FliDebugLevel)
        /// </summary>
        public string DebugFile
        {
            get { return _DebugFile; }
            set
            {
                _DebugFile = value;
            }
        }
        private string _DebugFile;

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable libflipro debug logging
        /// </summary>
        public bool DebugEnable
        {
            get { return _DebugEnable; }
            set
            {
                // If you are crashing here, make sure you are set to x64 and not debugging Any CPU
                ThrowOnError(NativeMethods.FPRODebug_EnableLevel(value, DebugLevel));
                _DebugEnable = value;
            }
        }
        private bool _DebugEnable = false;

        //-------------------------------------------
        /// <summary>
        /// Get/Set the libflipro debug log level
        /// </summary>
        public DBGLEVEL DebugLevel
        {
            get { return _DebugLevel; }
            set
            {
                ThrowOnError(NativeMethods.FPRODebug_EnableLevel(DebugEnable, value));
                _DebugLevel = value;
            }
        }
        private DBGLEVEL _DebugLevel = DBGLEVEL.NONE;

        #endregion

        //**************************************************************
        #region Public Classes

        public class FliFrameCountNameValueClass
        {
            public FliFrameCountNameValueClass() { }
            public FliFrameCountNameValueClass(string name, FliFrameCount value)
            {
                Name = name;
                Value = value;
            }

            public string Name { get; set; }
            public FliFrameCount Value { get; set; }

            public override string ToString()
            {
                return Name;
            }
        }

        public class FliFrameTypeNameValueClass
        {
            public FliFrameTypeNameValueClass() { }
            public FliFrameTypeNameValueClass(string name, FliFrameType value)
            {
                Name = name;
                Value = value;
            }

            public string Name { get; set; }
            public FliFrameType Value { get; set; }

            public override string ToString()
            {
                return Name;
            }
        }

        public class NameValuePair
        {
            public NameValuePair() { }
            public NameValuePair(string name, int value)
            {
                Name = name;
                Value = value;
            }

            public string Name { get; set; }
            public int Value { get; set; }

            public override string ToString()
            {
                return Name;
            }
        }

        public class SensorGain
        {
            public int Index { get; protected set; }
            public int DacIndex { get; protected set; }
            public double Gain
            {
                get { return _Gain; }
                set { _Gain = value; Name = _Gain.ToString("0.000"); }
            }
            private double _Gain = 0;
            public string Name { get; set; }

            public SensorGain()
            {
                Index = 0;
                DacIndex = 0;
                Gain = 0.0;
                Name = Gain.ToString("0.000");
            }
            public SensorGain(int index, double gain, int dacIndex)
            {
                Index = index;
                DacIndex = dacIndex;
                Gain = gain;
                Name = Gain.ToString("0.000");
            }
            public SensorGain(int index, uint gain, int dacIndex)
            {
                Index = index;
                DacIndex = dacIndex;
                Gain = (double)gain / GAIN_SCALE_FACTOR;
                Name = Gain.ToString("0.000");
            }

            public override string ToString()
            {
                return Name;
            }

        }

        public class BinRatio
        {
            public int Index { get; protected set; }
            public uint HBin { get; protected set; }
            public uint VBin { get; protected set; }
            public string Name { get { return ToString(); } }

            public BinRatio() : this(0,1,1) { }
            public BinRatio(int index, uint hbin, uint vbin)
            {
                Index = index;
                HBin = hbin;
                VBin = vbin;
            }

            public override string ToString()
            {
                return HBin.ToString() + ":" + VBin.ToString();
            }
        }

        public class WindowWrapper : System.Windows.Forms.IWin32Window
        {
            public WindowWrapper(IntPtr handle)
            {
                _hwnd = handle;
            }

            public WindowWrapper(Window window)
            {
                _hwnd = new WindowInteropHelper(window).Handle;
            }

            public IntPtr Handle
            {
                get { return _hwnd; }
            }

            private IntPtr _hwnd;
        }

        #endregion Public Classes

        //**************************************************************
        #region Observable Properties
        #endregion Observable Properties

        //**************************************************************
        #region Public Methods

        public void DebugLog(DBGLEVEL level, string message)
        {
            try
            {
                NativeMethods.FPRODebug_Write(level, message);
            }
            catch (Exception)
            {
                // ignore all exceptions
            }
        }
        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        /// <summary>
        /// Throw exception on native function error
        /// </summary>
        /// <param name="status"> return status from Native function (negative win32 error)</param>
        /// <param name="callerName"> name of calling method or property</param>
        protected void ThrowOnError(int status, [CallerMemberName]string callerName = "")
        {
            if (status < 0)
            {
                MessageBox.Show("Error encountered in: " + callerName + "\nStatus = " + status.ToString());
                //throw new FliException(-status, callerName);
            }
        }

        #endregion Protected Methods

        //**************************************************************
        #region Custom Exceptions

        //-------------------------------------------
        /// <summary>
        /// Exception thrown when native libfliPro functions
        /// return an error
        /// </summary>
        [Serializable]
        public class FliException : Win32Exception
        {
            public FliException(int error, string caller) :
                base(error)
            {
                Caller = caller;
            }
            public string Caller { get; private set; }

            public override string ToString()
            {
                string msg = "FliException <" + ErrorCode.ToString("X") + ">\n";
                msg += Message + "\n";
                msg += "From: " + Source + " (" + Caller + ")\n";
                msg += "Native Error: " + NativeErrorCode.ToString() + ", (" + (ErrorNumber)NativeErrorCode + ")\n";
                //msg += "Base Exception: " + GetBaseException().Message + "\n";
                msg += "StackTrace:\n";
                msg += StackTrace + "\n";
                return msg;
            }
        }

        public static string ExceptionToString(Exception ex)
        {
            string msg = "Exception \n";
            msg += ex.Message + "\n";
            msg += "From: " + ex.Source;
            msg += "Base Exception: " + ex.GetBaseException().Message + "\n";
            msg += "StackTrace:\n";
            msg += ex.StackTrace + "\n";
            return msg;
        }
        #endregion Custom Exceptions

        //**************************************************************
        #region IDisposable Support

        private bool isDisposed = false; // To detect redundant calls

        protected virtual void Dispose(bool disposing)
        {
            if (!isDisposed)
            {
                if (disposing)
                {
                    // Dispose managed state (managed objects).
                    OnDispose();
                }

                // TODO: free unmanaged resources (unmanaged objects) and override a finalizer below.
                // TODO: set large fields to null.
                isDisposed = true;
            }
        }

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
            // TODO: uncomment the following line if the finalizer is overridden above.
            // GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Perform any needed cleanup prior to disposing unmanaged objects
        /// </summary>
        protected virtual void OnDispose()
        {
        }

        #endregion IDisposable Support

    }

}
