﻿using System;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows;

namespace FliProCameraLib
{
    /// <summary>
    /// This file contains overridden methods specific 
    /// to first generation FLI Cameras using libfli.dll
    /// </summary>
    public class FliProCameraLegacy : FliProCamera
    {

        //**************************************************************
        #region Public Constants

        public const int FLI_DELAY_COUNTER_INTERVAL = 200;
        public const int FLI_MAX_SPEEDS = 8;
        public const int FLI_BGFLUSH_STOP = 0x0000;
        public const int FLI_BGFLUSH_START = 0x0001;
        public const int MAX_LEGACY_OVER_SCAN = 500;
        public const int FLI_EXPOSURE_TRIGGER_MASK = 0x0004;

        public const int MAX_EXPOSURE_TIME_MSECS = 39600000;       // 11 hours
        public const int MAX_EXPOSURE_TIME_SECS = 39600;           // 11 hours
        public const int MAX_EXPOSURE_TIME_MINS = 660;             // 11 hours
        public const int MAX_EXPOSURE_TIME_HRS = 11;               // 11 hours

        #endregion Public Constants

        //**************************************************************
        #region Constructor and Initialization

        public FliProCameraLegacy(FliDeviceId deviceId = new FliDeviceId())
        {
            Type = CameraType.GEN1;
            BitsPerPixel = FliFrame.BITS_PER_PIXEL_16;
            DeviceId = deviceId;
        }

        protected override void Initialize()
        {
            // Initialize The Frame Delay Timer
            FrameDelayTimer.Enabled = false;
            FrameDelayTimer.Elapsed += FrameDelayTimerEvent;
            FrameDelayTimer.AutoReset = true;

            // Initialize The Frame Delay Counter Timer
            FrameDelayCounterTimer.Enabled = false;
            FrameDelayCounterTimer.Elapsed += FrameDelayTimerEvent;
            FrameDelayCounterTimer.AutoReset = true;

            // Initialize The Grab Frame Timer
            GrabFrameTimer.Enabled = false;
            GrabFrameTimer.Elapsed += GrabFrameTimerEvent;
            GrabFrameTimer.AutoReset = true;

            // Initialize The Grab Frame Timer
            LEDTimer.Enabled = false;
            LEDTimer.Elapsed += LEDTimerEvent;
            LEDTimer.AutoReset = true;

            FliFrameCountList = new System.Collections.ObjectModel.ObservableCollection<FliFrameCountNameValueClass>();
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Single", FliFrameCount.SINGLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Multiple", FliFrameCount.MULTIPLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Continuous", FliFrameCount.CONTINUOUS));

            FliFrameTypeList = new System.Collections.ObjectModel.ObservableCollection<FliFrameTypeNameValueClass>();
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Normal", FliFrameType.NORMAL));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flat", FliFrameType.FLAT));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Dark", FliFrameType.DARK));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Flood", FliFrameType.FLOOD));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("RBI_Flush", FliFrameType.RBI_FLUSH));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Bias", FliFrameType.BIAS));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("TDI_Light", FliFrameType.TDI_LIGHT));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("TDI_Dark", FliFrameType.TDI_DARK));

            SupportedPixelDepths.Clear();
            SupportedPixelDepths.Add(16);

            //SupportedBinRatios.Add(new BinRatio(0, 1, 1));
            //SupportedBinRatios.Add(new BinRatio(1, 2, 2));
            //SupportedBinRatios.Add(new BinRatio(2, 2, 4));
            //BinRatioCount = 3;

            MaxHBin = 1;
            MaxVBin = 1;

        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Protected Properties

        protected int LEDpulsed { get; set; }
        protected int LEDRemain { get; set; }

        private const int PTCNRMBORDER = 20;	        // border around the main frame area that's skipped, and in the overscan Y direction
        private const int PTCOVSBORDER = 120;			// X border around the overscan regions that's skipped
        private const int PTCOVS = 500;			        // PTC overscan region width

        protected System.Timers.Timer FrameDelayTimer = new System.Timers.Timer();
        protected System.Timers.Timer FrameDelayCounterTimer = new System.Timers.Timer();
        protected System.Timers.Timer GrabFrameTimer = new System.Timers.Timer();
        protected System.Timers.Timer LEDTimer = new System.Timers.Timer();

        public FliDeviceId DeviceId { get; set; }
        private FliProCameraManager fliProCameraManager;

        #endregion Protected Properties

        //**************************************************************
        #region Observable Properties

        public override bool IsLegacyCamera { get { return true; } }

        #endregion Observable Properties

        //**************************************************************
        #region Public Properties

        public string Name { get { return DeviceId.Name; } }
        public DeviceDomain Domain { get { return DeviceId.Domain; } }

        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        /// <summary>
        /// Cancel the active exposure, if any
        /// </summary>
        public override void CancelExposure()
        {
            IsAborted = true;
            if (IsCapturing)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                long actualDelayTime = fliProCameraManager.ActiveCamera.GetAbsoluteExposureTime(fliProCameraManager.ActiveCamera.ExposureDelayTime,
                    fliProCameraManager.ActiveCamera.ExposureDelayUnits);
                if ((FrameCountType == FliFrameCount.CONTINUOUS) && VideoModeCapable && (actualDelayTime == 0))
                {
                    VideoModeEnabled = false;
                }
                else
                {
                    ThrowOnError(NativeMethods.FLICancelExposure(Handle));
                }
            }
        }

        /// <summary>
        /// Converts from Absolute time back to the display time in the given units
        /// </summary>
        /// <param name="value"></param>
        /// <param name="unit"></param>
        /// <returns></returns>
        public override int ConvertAbsoluteTimeToUnits(long value, ExposureUnits_Enum unit)
        {
            double factor = 1;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = (double)1 / 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = (long)1000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)1000 * 3600;
                    break;
            }
            value = (long) (value / factor);

            return (int)value;
        }

        public override void EndExposure()
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIEndExposure(Handle));
        }

        public override void ExposeFrame()
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            IsCapturing = true;
            IsAborted = false;

            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }

            // Only start the next Grab as video if this camera can do Video
            // And we are NOT using the Filter Wheel
            long actualDelayTime = GetAbsoluteExposureTime(ExposureDelayTime, ExposureDelayUnits);
            ThrowOnError(NativeMethods.FLIExposeFrame(Handle));

            GrabFrameTimer.Enabled = true;   // Need to turn the timer on to listen for completed data
        }

        public override long GetAbsoluteExposureTime(long value, ExposureUnits_Enum unit)
        {
            double factor = 1;
            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = (double)1 / 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = 1000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = 1000 * 3600;
                    break;
            }
            value = (long)(value * factor);

            return value;
        }

        public override void GetDeviceVersion(ref string firmwareVersion, ref string fpgaVersion, ref string controllerVersion)
        {
            try
            {
                StringBuilder version = new StringBuilder(100);
                NativeMethods.FLIGetLibVersion(version, (IntPtr)100);

                firmwareVersion = version.ToString();
                fpgaVersion = "N/A";
                controllerVersion = "N/A";
            }
            catch
            {
                firmwareVersion = "0.0.0.0";
                fpgaVersion = "N/A";
                controllerVersion = "N/A";
            }
        }

        public override uint GetFWRevision()
        {
            uint fwVersion = 0;
            NativeMethods.FLIGetFWRevision(Handle, out fwVersion);
            return fwVersion;
        }

        public override string GetLibVersion()
        {
            var version = new StringBuilder(MAX_STRING_LEN);
            if (Handle < 0)
            {
                IsAborted = true;
                return String.Empty;
            }
            ThrowOnError(NativeMethods.FLIGetLibVersion(version, (IntPtr)MAX_STRING_LEN));
            return version.ToString();
        }

        public override int GetRemainingExposureTime()
        {
            int time = 0;
            if (Handle < 0)
            {
                IsAborted = true;
                return time;
            }
            ThrowOnError(NativeMethods.FLIGetExposureStatus(Handle, out time));
            return time;
        }

        public override int GetTemperatures(ref double AmbientTemp, ref double BaseTemp, ref double CoolerTemp)
        {
            int status = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return -1;
            }
            AmbientTemp = 0;
            status = NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_CCD, out CoolerTemp);
            status = NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_BASE, out BaseTemp);
            status = NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_AMBIENT, out AmbientTemp);
            SensorTemperature = CoolerTemp;
            BaseTemperature = BaseTemp;
            AmbientTemperature = AmbientTemp;

            return status;
        }

        public override int GetCoolerDutyCycle(ref int DutyCycle)
        {
            int status = 0;
            double dutyCylceLocal = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return -1;
            }
            ThrowOnError(status = NativeMethods.FLIGetCoolerPower(Handle, out dutyCylceLocal));
            DutyCycle = (int)dutyCylceLocal;

            return status;
        }

        public override int ReadTemperature(int channel, ref double temp)
        {
            int status = 0;
            if (Handle < 0)
            {
                IsAborted = true;
                return -1;
            }
            ThrowOnError(status = NativeMethods.FLIReadTemperature(Handle, channel, out temp));
            return status;
        }

        public override void SetDefaultSupportedPixelDepths()
        {
            SupportedPixelDepths.Clear();
            SupportedPixelDepths.Add(16);
            PixelDepth = SupportedPixelDepths[0];
        }

        //-------------------------------------------
        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold the specified number of pixels at the current
        /// camera settings
        /// </summary>
        /// <param name="pixelcount">number of pixels</param>
        /// <returns></returns>
        public override int PixelsToBytes(int pixelcount)
        {
            return (((pixelcount * (int)PixelDepth) + 7) / 8);
        }

        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold an image of the specified size at the current
        /// camera settings
        /// </summary>
        /// <param name="width">image width in pixels</param>
        /// <param name="height">image height in pixels</param>
        /// <returns></returns>
        public override int PixelsToBytes(int width, int height)
        {
            return PixelsToBytes(width * height);
        }

        /// <summary>
        /// Start a Grab for this camera using the current parameter set
        /// </summary>
        public override void GrabFrame()
        {
            BeginGrabFrame();
        }

        /// <summary>
        /// Start a Grab for this camera using the current parameter set
        /// </summary>
        public override void GrabFrame(FliFrameCount frameCount)
        {
            BeginGrabFrame();
        }

        /// <summary>
        /// Grab one 8bpp frame from the camera
        /// </summary>
        /// <returns>Actual number of pixels grabbed</returns>
        public override uint GrabFrame(byte[,] buff)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return 0;
            }
            GCHandle BuffGch = GCHandle.Alloc(buff, GCHandleType.Pinned);
            IntPtr BuffPtr = BuffGch.AddrOfPinnedObject();
            IntPtr BuffLength = (IntPtr)(buff.Length * sizeof(byte));
            IntPtr BytesGrabbed = (IntPtr)0;

            try
            {
                ThrowOnError(NativeMethods.FLIGrabFrame(Handle, BuffPtr, BuffLength, out BytesGrabbed));
            }
            finally
            {
                BuffGch.Free();
            }
            return (uint)BytesGrabbed / sizeof(byte);
        }

        /// <summary>
        /// Grab one 16bpp frame from the camera
        /// </summary>
        /// <returns>number of pixels grabbed</returns>
        public override uint GrabFrame(ushort[,] buff)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return 0;
            }
            GCHandle BuffGch = GCHandle.Alloc(buff, GCHandleType.Pinned);
            IntPtr BuffPtr = BuffGch.AddrOfPinnedObject();
            IntPtr BuffLength = (IntPtr)(buff.Length * sizeof(ushort));
            IntPtr BytesGrabbed = (IntPtr)0;

            try
            {
               ThrowOnError(NativeMethods.FLIGrabFrame(Handle, BuffPtr, BuffLength, out BytesGrabbed));
            }
            finally
            {
                BuffGch.Free();
            }
            return (uint)BytesGrabbed / sizeof(ushort);
        }


        /// <summary>
        /// Grab one video frame from the camera
        /// Caller is responsible for pinning and freeing buffer
        /// </summary>
        public override void GrabVideoFrame(ushort[] BuffPtr, int BuffLength)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIGrabVideoFrame(Handle, BuffPtr, (IntPtr)BuffLength));
        }

        /// <summary>
        /// Grab one video frame from the camera
        /// </summary>
        public override void GrabVideoFrame(byte[] buff)
        {
            //if (Handle < 0)
            //{
            //    IsAborted = true;
            //    return;
            //}
            //GCHandle BuffGch = GCHandle.Alloc(buff, GCHandleType.Pinned);
            //IntPtr BuffPtr = BuffGch.AddrOfPinnedObject();
            //IntPtr BuffLength = (IntPtr)(buff.Length * sizeof(byte));

            //try
            //{
            //    ThrowOnError(NativeMethods.FLIGrabVideoFrame(Handle, BuffPtr, BuffLength));
            //}
            //finally
            //{
            //    BuffGch.Free();
            //}
        }

        /// <summary>
        /// Grab one 8bpp video frame from the camera
        /// </summary>
        public override void GrabVideoFrame(byte[,] buff)
        {
            //if (Handle < 0)
            //{
            //    IsAborted = true;
            //    return;
            //}
            //GCHandle BuffGch = GCHandle.Alloc(buff, GCHandleType.Pinned);
            //IntPtr BuffPtr = BuffGch.AddrOfPinnedObject();
            //IntPtr BuffLength = (IntPtr)(buff.Length * sizeof(byte));

            //try
            //{
            //    ThrowOnError(NativeMethods.FLIGrabVideoFrame(Handle, BuffPtr, BuffLength));
            //}
            //finally
            //{
            //    BuffGch.Free();
            //}
        }

        /// <summary>
        /// Grab one 16bpp frame from the camera
        /// </summary>
        public override void GrabVideoFrame(ushort[,] buff)
        {
            //if (Handle < 0)
            //{
            //    IsAborted = true;
            //    return;
            //}
            //GCHandle BuffGch = GCHandle.Alloc(buff, GCHandleType.Pinned);
            //IntPtr BuffPtr = BuffGch.AddrOfPinnedObject();
            //IntPtr BuffLength = (IntPtr)(buff.Length * sizeof(ushort));

            //try
            //{
            //    ThrowOnError(NativeMethods.FLIGrabVideoFrame(Handle, BuffPtr, BuffLength));
            //}
            //finally
            //{
            //    BuffGch.Free();
            //}
        }

        public override void SetVerticalTableEntry(int index, int height, int bin, int mode)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetVerticalTableEntry(Handle, index, height, bin, mode));
        }

        public override void GetVerticalTableEntry(int index, out int height, out int bin, out int mode)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                height = 0;
                bin = 1;
                mode = 0;
                return;
            }
            ThrowOnError(NativeMethods.FLIGetVerticalTableEntry(Handle, index, out height, out bin, out mode));
        }

        public override void EnableVerticalTable(int width, int offset, int flags)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIEnableVerticalTable(Handle, width, offset, flags));
        }

        //-------------------------------------------
        /// <summary>
        /// Read input port and raise NotifyPropertyChanged if needed
        /// </summary>
        /// <returns></returns>
        public override int ReadInputPort()
        {
            int value = 0;
            if (Handle < 0)
            {
                IsAborted = true;
                return value;
            }
            ThrowOnError(NativeMethods.FLIReadIOPort(Handle, out value));
            LastInputPort = value;
            return LastInputPort;

        }

        //-------------------------------------------
        /// <summary>
        /// Read camera sensor temperature 
        /// </summary>
        /// <returns></returns>
        public override double GetSensorTemperature()
        {
            double value = 0.0;
            if (Handle < 0)
            {
                IsAborted = true;
                return value;
            }
            ThrowOnError(NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_CCD, out value));
            SensorTemperature = value;
            return SensorTemperature;
        }

        //-------------------------------------------
        /// <summary>
        /// Read camera base temperature 
        /// </summary>
        /// <returns></returns>
        public override double GetBaseTemperature()
        {
            double value = 0.0;
            if (Handle < 0)
            {
                IsAborted = true;
                return value;
            }
            ThrowOnError(NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_BASE, out value));
            BaseTemperature = value;
            return BaseTemperature;
        }

        public override int ReadThumbnail(long length, byte[] rbuf)
        {
            bool oldEnabled = TemperatureStatusTimer.Enabled;
            EnableTemperatureTimer(false);
            int status = NativeMethods.FLIReadThumbnail(Handle, length, rbuf);
            EnableTemperatureTimer(oldEnabled);
            return status;
        }

        public override int WriteThumbnail(long length, byte[] wbuf)
        {
            bool oldEnabled = TemperatureStatusTimer.Enabled;
            EnableTemperatureTimer(false);
            int status = NativeMethods.FLIWriteThumbnail(Handle, length, wbuf);
            EnableTemperatureTimer(oldEnabled);
            return status;
        }


        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        protected void ControlBackgroundFlush(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIControlBackgroundFlush(Handle, value));
        }

        protected void LockDevice(bool value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            if (value)
            {
                NativeMethods.FLILockDevice(Handle);
            }
            else
            {
                NativeMethods.FLIUnlockDevice(Handle);
            }
        }

        protected override void SetBinningIndex(int value)
        {
            for (int i = 0; i < SupportedBinRatios.Count; i++)
            {
                if (SupportedBinRatios[i].Index == value)
                {
                    VBin = SupportedBinRatios[i].VBin;
                    HBin = SupportedBinRatios[i].HBin;
                    break;
                }
            }
        }

        protected override void SetFrameInterval(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            // ThrowOnError(NativeMethods.???(Handle));
        }

        protected override void SetTargetTemperature(double value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            NativeMethods.FLISetTemperature(Handle, value);
        }

        protected override void SetPixelDepth(uint value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            //ThrowOnError(NativeMethods.FLISetBitDepth(Handle, (int)value));
        }

        //-------------------------------------------
        /// <summary>
        /// Open/Close Shutter
        /// </summary>
        /// <param name="value"></param>
        public override void SetShutterOpen(bool open)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            if (open)
            {
                ThrowOnError(NativeMethods.FLIControlShutter(Handle, FliShutterControl.OPEN.GetHashCode()));
            }
            else
            {
                ThrowOnError(NativeMethods.FLIControlShutter(Handle, FliShutterControl.CLOSE.GetHashCode()));
            }
        }

        protected override void SetShutterDelay(int opendelay, int closedelay)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            //ThrowOnError(NativeMethods.FPROCtrl_SetShutterDelay(Handle, (ushort)opendelay, (ushort)closedelay));
        }

        protected override void SetTDIRate(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetTDI(Handle, value, 0));
        }

        protected override void SetMode(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetCameraMode(Handle, value));
        }

        protected override void SetHeaterEnable(bool value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            //FIXME: Need to find the heater function in the old camera
            //ThrowOnError(NativeMethods.FPROCtrl_SetHeaterConfig(Handle, value, (byte)_HeaterPower));
        }

        protected override void SetBinning()
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            // All binning is done in software so send 1
            //ThrowOnError(NativeMethods.FLISetHBin(Handle, (int)HBin));
            //ThrowOnError(NativeMethods.FLISetVBin(Handle, (int)VBin));
            ThrowOnError(NativeMethods.FLISetHBin(Handle, 1));
            ThrowOnError(NativeMethods.FLISetVBin(Handle, 1));
        }

        protected override void SetFrameType(FliFrameType value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }

            if (value == FliFrameType.DARK)
            {
                // For Bias we want to set the FrameType to NORMAL and the Exposure Time to 0
                ThrowOnError(NativeMethods.FLISetFrameType(Handle, FliFrameType.NORMAL.GetHashCode()));
                ThrowOnError(NativeMethods.FLISetExposureTime(Handle, 0));
            }
            else
            {
                // NOTE: if the frame type is FOCUS or MULTIPLE we just want to send down NORMAL
                ThrowOnError(NativeMethods.FLISetFrameType(Handle, value.GetHashCode()));
                SetExposure(ExposureTime, ExposureDelayTime, ExposureUnits, ExposureDelayUnits, Immediate);
            }
        }

        public override void SetFrameTypeForced(FliFrameType value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }

            if (value == FliFrameType.DARK)
            {
                // For Bias we want to set the FrameType to NORMAL
                ThrowOnError(NativeMethods.FLISetFrameType(Handle, FliFrameType.DARK.GetHashCode()));
            }
            else
            {
                // For Bias we want to set the FrameType to NORMAL
                ThrowOnError(NativeMethods.FLISetFrameType(Handle, FliFrameType.NORMAL.GetHashCode()));
            }
        }

        protected override void SetNFlushes(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetNFlushes(Handle, value));
        }

        protected override void SetExposure(long value, long delay, ExposureUnits_Enum unit, ExposureUnits_Enum unitDelay, bool Immediate)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }

            value = GetAbsoluteExposureTime(value, unit);

            ThrowOnError(NativeMethods.FLISetExposureTime(Handle, (int)value));
        }

        protected override void SetShutter(FliShutterControl value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            if (value < FliShutterControl.EXPOSURE_CONTROL)
            {
                int rtn = NativeMethods.FLIControlShutter(Handle, (int)value);
                ThrowOnError(rtn);
            }
        }


        public override FliRectangle GetBaseExposureArea()
        {
            FliRectangle rec = new FliRectangle(
                ImageArea.UL.X + PTCNRMBORDER,
                ImageArea.UL.Y + PTCNRMBORDER,
                ImageArea.LR.X - PTCNRMBORDER,
                ImageArea.LR.Y - 2 - PTCNRMBORDER);
            return rec;
        }
        
        //-------------------------------------------
        /// <summary>
        /// Get the description string for the specified camera mode index
        /// </summary>
        /// <param name="mode">camera mode index</param>
        /// <returns>string or "" if mode index is not supported</returns>
        public override string GetCameraModeString(int mode)
        {
            var str = new StringBuilder(256);
            if (0 != NativeMethods.FLIGetCameraModeString(Handle, mode, str, (IntPtr)256))
            {
                return "";
            }
            return str.ToString();
        }

        public override FliRectangleWH GetReadoutDimensions()
        {
            var rect = new FliRectangleWH();
            if (Handle < 0)
            {
                IsAborted = true;
                return rect;
            }
            int hbin;
            int vbin;
            ThrowOnError(NativeMethods.FLIGetReadoutDimensions(Handle,
                                        out rect.WH.W, out rect.UL.X, out hbin,
                                        out rect.WH.H, out rect.UL.Y, out vbin));

            if (SetValue<FliRectangleWH>(ref _LastReadoutAreaWH, rect, "LastReadoutAreaWH"))
            {
                OnPropertyChanged("LastReadoutArea");
            }
            HBin = (uint)hbin;
            VBin = (uint)vbin;
            return rect;
        }

        protected override void SetDac(uint value)
        {
            //-=-= WRITEME
        }

        protected override void SetPortConfig(int value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIConfigureIOPort(Handle, value));
        }

        protected override void SetVideoModeEnabled(bool value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }

            try
            {
                int status = -1;
                if (value)
                {
                    // Configure the Exposure External Trigger port to allow software triggers
                    ThrowOnError(NativeMethods.FLIConfigureIOPort(Handle, FLI_EXPOSURE_TRIGGER_MASK));
                    status = NativeMethods.FLIStartVideoMode(Handle);
                }
                else
                {
                    // Configure the Exposure External Trigger port to to off
                    ThrowOnError(NativeMethods.FLIConfigureIOPort(Handle, 0));
                    status = NativeMethods.FLIStopVideoMode(Handle);
                }
            }
            catch(Exception ex)
            {
                MessageBox.Show("Error encountered calling FliProCameraLegacy.SetVideoModeEnabled: " + ex.Message, "Warning", MessageBoxButton.OK,
                                MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        protected override void SetFanEnable(bool value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetFanSpeed(Handle, (value ? (int)FliFanSpeed.ON : (int)FliFanSpeed.OFF)));
        }

        // OBSOLETE
        //protected override void SetFanPower(uint value)
        //{
        //    int speed = (int)FliFanSpeed.OFF;
        //    if (FanIsEnabled && value > 0)
        //    {
        //        speed = (int)FliFanSpeed.ON;
        //    }
        //    ThrowOnError(NativeMethods.FLISetFanSpeed(Handle, speed));
        //}

        protected override int OpenDevice()
        {
            return NativeMethods.FLIOpen(out _Handle, Name, (int)Domain);
        }

        protected override void CloseDevice()
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLIClose(Handle));
        }

        //-------------------------------------------
        /// <summary>
        /// Update dynamically changing values
        /// </summary>
        protected override void UpdateStatus()
        {
            GetBaseTemperature();
            GetSensorTemperature();
            // Add other camera status updates here
        }


        //-------------------------------------------
        /// <summary>
        /// Device Specific actions when connection is established
        /// </summary>
        protected override void OnConnect()
        {
            // Make sure any defaults get down to the camera
            ForcePropertyChange = true;

            var sernum = new StringBuilder(MAX_STRING_LEN);
            NativeMethods.FLIGetSerialString(Handle, sernum, (IntPtr)sernum.Capacity);
            SerialNumber = sernum.ToString();

            int hwrev = 0;
            NativeMethods.FLIGetHWRevision(Handle, out hwrev);
            HwVersion = hwrev;

            int fwrev = 0;
            NativeMethods.FLIGetHWRevision(Handle, out fwrev);
            FwVersion = fwrev;

            var model = new StringBuilder(MAX_STRING_LEN);
            NativeMethods.FLIGetModel(Handle, model, (IntPtr)MAX_STRING_LEN);
            Model = model.ToString();

            var array = new FliRectangle();
            NativeMethods.FLIGetArrayArea(Handle, out array.UL.X, out array.UL.Y, out array.LR.X, out array.LR.Y);
            ArrayArea = array;
            MaxImageWidth = ArrayAreaWH.W + MAX_LEGACY_OVER_SCAN;
            MaxImageHeight = ArrayAreaWH.H + MAX_LEGACY_OVER_SCAN;

            VisibleArea = GetVisibleArea();

            var pixel = new FliPixelSize();
            NativeMethods.FLIGetPixelSize(Handle, out pixel.W, out pixel.H);
            pixel.W *= 1000000.0;
            pixel.H *= 1000000.0;
            PixelSize = pixel;

            int pattern;
            NativeMethods.FLIGetCameraColorPattern(Handle, out pattern);
            ColorPattern = (FliBayerPattern)pattern;

            double temp;
            int status = NativeMethods.FLIReadTemperature(Handle, FLI_TEMPERATURE_CCD, out temp);

            // Put the camera in a known state
            Shutter = FliShutterControl.CLOSE;
            TDIRate = 0;
            //BinRatioValue = SupportedBinRatios[0];
            BinRatioValue = new BinRatio(0, 1, 1);

            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            NFlushes = 1;
            ImageArea = VisibleArea;
            VideoModeEnabled = false;

            // Camera Controls
            ExposureTime = Exposure_Enum.exposure_600.GetHashCode();
            ShutterOpen = false;
            IlluminationOn = false;
            Immediate = false;
            LEDOn = false;
            HeaterEnabledOn = false;
            FanIsEnabledOn = true;
            PixelDepth = SupportedPixelDepths[0];
            HorizontalScanDirInverted = false;
            VerticalScanDirInverted = false;
            SupportedModes.Clear();

            // Initialize supported modes
            for (int i = 0; i < FLI_MAX_SPEEDS; i++)
            {
                string modeStr = GetCameraModeString(i);
                if (String.IsNullOrEmpty(modeStr))
                {
                    break;
                }
                else
                {
                    SupportedModes.Add(new NameValuePair(modeStr, i));
                }
            }
            ModeCount = (uint)SupportedModes.Count;

            int mode;
            NativeMethods.FLIGetCameraMode(Handle, out mode);
            Mode = mode;

            HighGainTable.Clear();
            LowGainTable.Clear();
            HighGainTable.Add(new SensorGain(1, 1500, 0));
            HighGainValue = HighGainTable[0];
            LowGainTable.Add(new SensorGain(1, 1500, 0));
            LowGainValue = LowGainTable[0];

            // Normalize the active area to begin at (0, 0)
            int newBottom = VisibleArea.LR.Y;
            int newRight = VisibleArea.LR.X;
            if (VisibleArea.UL.X > 0) { newRight = VisibleArea.LR.X - VisibleArea.UL.X; }
            if (VisibleArea.UL.Y > 0) { newBottom = VisibleArea.LR.Y - VisibleArea.UL.Y; }
            VisibleArea = new FliRectangle(0, 0, newRight, newBottom);

            PortConfig = 0;    // make all signals inputs
            ControlBackgroundFlush(FLI_BGFLUSH_START);

            // See if this camera can perform Video Capture
            int results = NativeMethods.FLIStopVideoMode(Handle);
            VideoModeCapable = (results != NativeMethods.INVALID_ARGUMENT);

            // Now only send changes down to the camera
            ForcePropertyChange = false;
        }

        public override FliRectangle GetSensorArea()
        {
            var visible = new FliRectangle();
            NativeMethods.FLIGetArrayArea(Handle, out visible.UL.X, out visible.UL.Y, out visible.LR.X, out visible.LR.Y);
            visible.LR.X -= visible.UL.X;
            visible.LR.Y -= visible.UL.Y;
            visible.UL.X = 0;
            visible.UL.Y = 0;
            return visible;
        }

        public override FliRectangle GetVisibleArea()
        {
            var visible = new FliRectangle();
            NativeMethods.FLIGetVisibleArea(Handle, out visible.UL.X, out visible.UL.Y, out visible.LR.X, out visible.LR.Y);
            visible.LR.X -= visible.UL.X;
            visible.LR.Y -= visible.UL.Y;
            visible.UL.X = 0;
            visible.UL.Y = 0;
            return visible;
        }

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="frameCount">number of frames to capture. 0=no limit</param>
        /// <returns>false on any error</returns>
        protected override bool SetStartCapture(int frameCount)
        {
            bool ok = false;
            int result = NativeMethods.FPROFrame_CaptureStart(Handle, (uint)frameCount);
            ok = (result >= 0);
            return ok;
        }

        //-------------------------------------------
        /// <summary>
        /// Stop current frame capture sequence, if any)
        /// </summary>
        protected override void SetStopCapture()
        {
            NativeMethods.FPROFrame_CaptureStop(Handle);
        }


        protected override void SetImageArea(FliRectangle value)
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            ThrowOnError(NativeMethods.FLISetImageArea(Handle,
                value.UL.X, value.UL.Y, value.LR.X, value.LR.Y));
        }

        protected override void SetImageArea(FliRectangleWH value)
        {
            RequestedImageAreaWH = new FliRectangleWH(value.UL.X, value.UL.Y, value.W, value.H);
            SetImageArea(new FliRectangle(value));
        }

        protected override uint GetRawImageFrame(byte[] framebuf, int framesize, bool useDelay)
        {
            uint size = (uint)framesize;
            //-=-= TODO: Write replacement for GrabVideoFrame, remove deprecated Grab*
            return size;
        }

        #endregion Protected Methods

        //**************************************************************
        #region Private Methods

        private void BeginGrabFrame()
        {
            if (Handle < 0)
            {
                //CameraError();
                //EndGrabFrame();
                IsAborted = true;
                return;
            }

            //if (m_bExternalTrigger == TRUE)
            //{
            //    BOOL ExtTriggerLevel = ReadProfileInt("Grab", "ExternalTriggerLevel", FALSE);
            //    if (ExtTriggerLevel == FALSE)
            //    {
            //        FLIControlShutter(dev, FLI_SHUTTER_EXTERNAL_TRIGGER_LOW);
            //    }
            //    else
            //    {
            //        FLIControlShutter(dev, FLI_SHUTTER_EXTERNAL_TRIGGER_HIGH);
            //    }
            //}

            // Original
            //var visible = new FliRectangle();
            //NativeMethods.FLIGetVisibleArea(Handle, out visible.UL.X, out visible.UL.Y, out visible.LR.X, out visible.LR.Y);
            //FliPoint Offset = new FliPoint(visible.UL.X + ImageArea.UL.X, visible.UL.Y + ImageArea.UL.Y);
            ////FliPoint Offset = new FliPoint(visible.UL.X, visible.UL.Y);
            //FliPixelSize FrameSize = new FliPixelSize(ImageAreaWH.W / HBin, ImageAreaWH.H / VBin);

            //SetFrameType(FrameType);
            //SetFrameTypeForced(FrameType);
            //SetBinning();
            //SetImageArea(new FliRectangle(Offset.X, Offset.Y, Offset.X + Convert.ToInt32(FrameSize.W), Offset.Y + Convert.ToInt32(FrameSize.H)));

            var visible = new FliRectangle();
            NativeMethods.FLIGetVisibleArea(Handle, out visible.UL.X, out visible.UL.Y, out visible.LR.X, out visible.LR.Y);
            FliRectangle arrayArea = new FliRectangle();
            NativeMethods.FLIGetArrayArea(Handle, out arrayArea.UL.X, out arrayArea.UL.Y, out arrayArea.LR.X, out arrayArea.LR.Y);
            FliRectangleWH arrayAreaWH = new FliRectangleWH(arrayArea);

            // Start by using the visual area of the sensor
            FliRectangleWH sensorArea = new FliRectangleWH(visible.UL.X + ImageArea.UL.X, visible.UL.Y + ImageArea.UL.Y, ImageAreaWH.W, ImageAreaWH.H);

            if (UseOverscanColumns)
            {
                // If they are using the Overscan columns then Interpret the ImageArea origin relative to Total Sensor area
                sensorArea.UL = ImageArea.UL;
            }

            // Add the current ImageAreaWH to modify the data received from the camera if necessary 
            TargetImageAreaWHList.Clear();
            // Set this here so that when we receive the data from the camera we know what we're getting
            //ImageAreaWH = sensorArea;

            SetFrameType(FrameType);
            SetFrameTypeForced(FrameType);
            SetBinning();
            SetImageArea(sensorArea);
            SetExposure(ExposureTime, ExposureDelayTime, ExposureUnits, ExposureDelayUnits, Immediate);

            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }

            // Only start the next Grab as video if this camera can do Video
            // And we are NOT using the Filter Wheel
            long actualDelayTime = fliProCameraManager.ActiveCamera.GetAbsoluteExposureTime(fliProCameraManager.ActiveCamera.ExposureDelayTime,
                fliProCameraManager.ActiveCamera.ExposureDelayUnits);
            GrabFrameTimer.Interval = 5;

            SetNFlushes(0);
            ExposeFrame();
        }

        private void CameraError()
        {
            PublishStatus("Comm error with Camera", false);
            CloseDevice();
        }

        private void EndGrabFrame()
        {
            LockDevice(false);

            GrabFrameTimer.Enabled = false;
            FrameDelayTimer.Enabled = false;
            FrameDelayCounterTimer.Enabled = false;

            IsCapturing = false;
            //SetTimer(EVENT_CAMERASTATUS, 2500, NULL);

            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }

            fliProCameraManager.StatusProgressFinished();

            PublishStatus("Stopped", true);

            if (BackgroundFlush == FliBgFlush.START)
            {
                ControlBackgroundFlush(FLI_BGFLUSH_START);
            }
        }

        private bool GrabDataFromCameraFrame()
        {
            FliFrame fliFrame = new FliFrame(this);
            int currentRow = 0;
            bool continueGrab = false;

            //StatusProgressIncrement();
            fliFrame.CameraID = SerialNumber;
            fliFrame.BitsPerPixel = FliFrame.BITS_PER_PIXEL_16;
            //fliFrame.SlopeHighGain = imageParams.SlopeHighGain;
            //fliFrame.SlopeLowGain = imageParams.SlopeLowGain;
            //fliFrame.BlackLevelOffsetHigh = imageParams.BlackLevelOffsetHigh;
            //fliFrame.BlackLevelOffsetLow = imageParams.BlackLevelOffsetLow;
            //fliFrame.HighLowThreshold = imageParams.HighLowThreshold;
            fliFrame.FrameAreaWH = new FliRectangleWH(RequestedImageAreaWH.UL.X, RequestedImageAreaWH.UL.Y, RequestedImageAreaWH.W, RequestedImageAreaWH.H);
            fliFrame.FrameAreaWHTotal = fliFrame.FrameAreaWH;
            int totalPixels = fliFrame.FrameAreaWH.W * fliFrame.FrameAreaWH.H;

            // Allocate the Metadata
            FliProFrameMetadata fliProFrameMetadata = new FliProFrameMetadata();
            int lineSize = Convert.ToInt32(fliFrame.FrameAreaWH.W * sizeof(ushort));   // There are 2 bytes per pixel
            int byteSize = lineSize * fliFrame.FrameAreaWH.H;

            fliFrame.RawDataLength = (uint)(byteSize + fliProFrameMetadata.MetadataLength);
            fliProFrameMetadata.SensorPixelBitDepth = FliFrame.BITS_PER_PIXEL_16.GetHashCode();
            fliProFrameMetadata.DataPixelBitDepth = FliFrame.BITS_PER_PIXEL_16.GetHashCode();
            fliProFrameMetadata.DummyPixelsPerRow = 0;
            fliProFrameMetadata.HDRMode = false;
            fliProFrameMetadata.SetPaddedModel(Model);
            fliProFrameMetadata.SetPaddedSerialNumber(SerialNumber);
            fliProFrameMetadata.HorizontalPixels = fliFrame.FrameAreaWH.W;
            fliProFrameMetadata.VerticalPixels = fliFrame.FrameAreaWH.H;
            fliProFrameMetadata.PreDummyRows = 0;
            fliProFrameMetadata.PostDummyRows = 0;
            fliProFrameMetadata.ExposureTime = ExposureTime;
            fliProFrameMetadata.SensorColdTemp = Convert.ToInt32(BaseTemperature);
            fliProFrameMetadata.SensorTemp = Convert.ToInt32(Temperature);
            fliProFrameMetadata.HDRMode = false;
            fliProFrameMetadata.HorizontalBinning = 1;
            fliProFrameMetadata.VerticalBinning = 1;
            fliProFrameMetadata.CaptureString = DateTime.Now.ToString("yyyy-MM-ddTHH:mm:ss.fff000");
            fliProFrameMetadata.LowGain = 0;
            fliProFrameMetadata.HighGain = 0;

            // Remember Rows are the Y axis and Columns are the X axis
            fliProFrameMetadata.ImageStartRow = RequestedImageAreaWH.UL.Y;
            fliProFrameMetadata.ImageStartColumn = RequestedImageAreaWH.UL.X;
            fliProFrameMetadata.ImageStopRow = RequestedImageAreaWH.UL.Y + RequestedImageAreaWH.H - 1;
            fliProFrameMetadata.ImageStopColumn = RequestedImageAreaWH.UL.X + RequestedImageAreaWH.W - 1;

            fliProFrameMetadata.FrameNumber = 1;
            uint fwVersion = 0;
            NativeMethods.FLIGetFWRevision(Handle, out fwVersion);
            fliProFrameMetadata.FirmwareVersion = fwVersion;

            fliFrame.FrameMetaData = fliProFrameMetadata;

            // Only start the next Grab as video if this camera can do Video
            // And we are NOT using the Filter Wheel
            long actualDelayTime = fliProCameraManager.ActiveCamera.GetAbsoluteExposureTime(fliProCameraManager.ActiveCamera.ExposureDelayTime,
                fliProCameraManager.ActiveCamera.ExposureDelayUnits);
                ushort[] rowData = new ushort[fliFrame.FrameAreaWH.W];
                fliProCameraManager.StatusProgressInitialize();
                fliProCameraManager.SetStatusProgress(fliFrame.FrameAreaWH.H);
                for (int i = 0; i < fliFrame.FrameAreaWH.H; i++)
                {
                    if (IsAborted == true)
                    {
                        EndGrabFrame();
                        return continueGrab;
                    }
                    if (NativeMethods.FLIGrabRow(Handle, rowData, (IntPtr)fliFrame.FrameAreaWH.W) != 0)
                    {
                        //EndGrabFrame();
                        return continueGrab;
                    }
                    Array.Copy(rowData, 0, fliFrame.RawData, currentRow, fliFrame.FrameAreaWH.W);
                    currentRow += fliFrame.FrameAreaWH.W;
                    fliProCameraManager.StatusProgressIncrement(true);
                }

            fliFrame.IsValid = true;

            LastFrameQueued = fliFrame;
            return continueGrab;
        }

        #endregion Private Methods

        //**************************************************************
        #region Timer Event Methods

        private void FrameDelayTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            FrameDelayTimer.Enabled = false;
            FrameDelayCounterTimer.Enabled = false;
            if (IsAborted == true)
            {
                EndGrabFrame();
                return;
            }
            GrabFrame();
        }

        private void FrameDelayCounterTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            DelayCounter += FLI_DELAY_COUNTER_INTERVAL;
            if (IsAborted == true)
            {
                EndGrabFrame();
                return;
            }
            PublishStatus("Delaying (" + Convert.ToUInt64(((DelayCounter / 1000) / FrameDelay) * 100) + ") ", true);
        }

        private void CameraStatusTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
        }

        private void GrabFrameTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            // Turn off the timer so that it doesn't fire while we are processing
            // We'll turn it back on at the end if we're not done
            GrabFrameTimer.Enabled = false;   
            int status = 0;
            int exposureLeft = 0;
            int camera_status = 0;
            double exposureLeftDbl = 0;
            double hrs, mins, secs;
            int LEDpulsed = 0;

            if (Handle < 0)
            {
                EndGrabFrame();
                return;
            }

            if (IsAborted)
            {
                EndGrabFrame();
                return;
            }

            /* This will only fail if the camera is offline */
            ThrowOnError(NativeMethods.FLIGetDeviceStatus(Handle, out camera_status));

            if (FrameCountType != FliFrameCount.CONTINUOUS)
            {
                ThrowOnError(status = NativeMethods.FLIGetExposureStatus(Handle, out exposureLeft));

                if (status != 0)
                {
                    EndGrabFrame();
                    CameraError();
                    return;
                }
            }

            switch (camera_status & 0x03)
            {
                case 0x00:
                    PublishStatus("Camera Idle", true);
                    break;

                case 0x01:
                    PublishStatus("Waiting for trigger", true);
                    break;

                case 0x02:
                    exposureLeftDbl = exposureLeft / 1000.0;
                    hrs = Math.Floor((exposureLeftDbl / (60.0 * 60.0)));
                    mins = Math.Floor((exposureLeftDbl - (hrs * 60.0 * 60.0)) / 60.0);
                    secs = exposureLeftDbl - (hrs * 60.0 * 60.0) - (mins * 60.0);

                    if (fliProCameraManager == null)
                    {
                        fliProCameraManager = FliProCameraManager.Instance;
                    }

                    // Always show the remaining time in the bottom status bar
                    fliProCameraManager.ExposureStatusText = hrs.ToString("00") + ":" + mins.ToString("00") + ":" + secs.ToString("00");

                    // Only show it at the top if the capture type is a single frame
                    if ((FrameCountType != FliFrameCount.CONTINUOUS) && (FrameCountType != FliFrameCount.MULTIPLE))
                    {
                        fliProCameraManager.ProgressTimerClock = fliProCameraManager.ExposureStatusText;
                    }

                    if ((LEDRemain > 0) && (LEDOnTime > 0) && (LEDpulsed == 0))
                    {
                        LEDRemain = LEDOnTime;
                        LEDTimer.Interval = LEDRemain;
                        LEDpulsed = 1;
                    }

                    break;

                case 0x03:
                    PublishStatus("Reading", true);
                    break;

                default:
                    break;
            }

            if ((exposureLeft == 0) && ((camera_status & 0x80000000) == 0x80000000))
            {
                GrabFrameTimer.Enabled = GrabDataFromCameraFrame();
                if (!GrabFrameTimer.Enabled)
                {
                    IsCapturing = false;
                    LEDpulsed = 0;
                }
            }
            else
            {
                // Turn the timer back on
                GrabFrameTimer.Enabled = true;
            }
        }
        
        private void LEDTimerEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            if (LEDRemain > 0)
            {
                LEDTimer.Enabled = false;
                LEDRemain = 0;
            }
        }

        #endregion Timer Event Methods
    }
}
